from ribosome.machine import message, json_message

StageI = message('StageI')
StageII = message('StageII')
AddPlugin = json_message('AddPlugin', 'spec')
ShowPlugins = message('ShowPlugins')
SetupPlugins = message('SetupPlugins')
SetupVenvs = message('SetupVenvs')
PostSetup = message('PostSetup')
InstallMissing = message('InstallMissing')
AddVenv = message('AddVenv', 'venv')
IsInstalled = message('Installed', 'venv')
Installed = message('Installed', 'venv')
Updated = message('Updated', 'venv')
Activate = message('Activate', 'name', varargs='plugins')
Activated = message('Activated', 'venv')
UpdatePlugins = message('UpdatePlugins', varargs='plugins')
Reboot = message('Reboot', 'venv')

__all__ = ('StageI', 'AddPlugin', 'ShowPlugins', 'SetupPlugins', 'SetupVenvs', 'InstallMissing', 'AddVenv',
           'IsInstalled', 'Installed', 'Updated', 'Activate', 'Activated', 'StageII', 'PostSetup', 'UpdatePlugins',
           'Reboot')



from types import SimpleNamespace

from amino import Path, List, Try, Either
from amino.util.string import ToStr

from chromatin.plugin import VimPlugin


class Venv(ToStr):

    def __init__(self, dir: Path, ns: SimpleNamespace, plugin: VimPlugin) -> None:
        self.dir = dir
        self.ns = ns
        self.plugin = plugin

    def _arg_desc(self) -> List[str]:
        return List(str(self.dir), str(self.plugin))

    @property
    def req(self) -> str:
        return self.plugin.spec

    @property
    def name(self) -> str:
        return self.plugin.name

    @property
    def site(self) -> Path:
        return self.dir / 'lib' / 'python3.6' / 'site-packages'

    @property
    def plugin_path(self) -> Path:
        return self.site / self.name / 'nvim_plugin.py'

    @property
    def python_executable(self) -> Either[str, Path]:
        return Try(lambda: self.ns.env_exe) / Path

__all__ = ('Venv',)



from ribosome.nvim import NvimFacade

from chromatin.venv import Venv
from chromatin.logging import Logging

from amino import Either, Path, do

host_start_function_name = 'CrmStartHost'

host_start_function_code = f'''
    function! {host_start_function_name}(args, host)
        let channel_id = jobstart(a:args, {{'rpc': v:true}})
        if rpcrequest(channel_id, 'poll') ==# 'ok'
            return channel_id
        else
            throw 'could not load host ' . a:host.name
        endif
    endfunction
'''

require_function_name = 'remote#host#Require'


def start_host_cmd(name: str, python_exe: Path, plug: Path) -> str:
    cmdline = [str(python_exe), '-c', 'import neovim; neovim.start_host()', str(plug)]
    return f'''call remote#host#Register('{name}', '*', function('{host_start_function_name}', [{cmdline!r}]))'''


class PluginHost(Logging):

    def __init__(self, vim: NvimFacade) -> None:
        self.vim = vim

    @do
    def start(self, venv: Venv) -> Either[str, str]:
        if not self.vim.function_exists(host_start_function_name):
            yield self.vim.execute(host_start_function_code)
        exe = yield venv.python_executable
        yield self.vim.cmd_sync(start_host_cmd(venv.name, exe, venv.plugin_path))

    def require(self, venv: Venv) -> Either[str, str]:
        return self.vim.call(require_function_name, venv.name)


__all__ = ('PluginHost',)



from amino import Path, List

xdg_cache_home_env_var = 'XDG_CACHE_HOME'


def create_venv_dir_error(dir: Path) -> str:
    return f'could not create venv dir `{dir}`'


def no_plugins_match_for_update(names: List[str]) -> str:
    return f'no plugins match for update: {names.join_comma}'


def installed_plugin(name: str) -> str:
    return f'installed plugin \'{name}\''


def updated_plugin(name: str) -> str:
    return f'updated plugin \'{name}\''

__all__ = ('xdg_cache_home_env_var', 'create_venv_dir_error', 'installed_plugin', 'updated_plugin')



from typing import Any

from ribosome.data import Data
from ribosome.record import dfield, list_field, map_field, field
from ribosome.nvim import NvimFacade, AsyncVimProxy

from amino import List, _, Either, Path, Try, env, Right, do, Boolean
from amino.boolean import true

from chromatin.logging import Logging
from chromatin.plugin import VimPlugin
from chromatin.venv import Venv
from chromatin.venvs import VenvFacade
from chromatin.util.resources import xdg_cache_home_env_var, create_venv_dir_error


@do
def _default_venv_dir() -> Either[str, Path]:
    xdg_cache_path = env[xdg_cache_home_env_var] / Path | (Path.home() / '.cache')
    venv_dir = xdg_cache_path / 'chromatin' / 'venvs'
    yield Try(venv_dir.mkdir, parents=True, exist_ok=True).lmap(lambda a: create_venv_dir_error(venv_dir))
    yield Right(venv_dir)


class Env(Data, Logging):
    vim = field((NvimFacade, AsyncVimProxy))
    initialized = dfield(False)
    plugins = list_field(VimPlugin)
    venvs = map_field()
    installed = list_field(Venv)
    active = list_field(Venv)

    def add_plugin(self, name: str, spec: str) -> 'Env':
        return self.append1.plugins(VimPlugin(name=name, spec=spec))

    @property
    def show_plugins(self) -> List[str]:
        def format(plug: VimPlugin) -> str:
            return plug.spec
        return self.plugins.map(format).cons('Configured plugins:')

    def add_venv(self, venv: Venv) -> 'Env':
        return self.modder.venvs(_ + (venv.plugin, venv))

    def add_installed(self, venv: Venv) -> 'Env':
        return self.append1.installed(venv)

    def missing_in(self, venvs: VenvFacade) -> List[Venv]:
        return self.venvs.v.filter_not(venvs.package_installed)

    @property
    @do
    def missing(self) -> Either[str, List[Venv]]:
        venv_facade = yield self.venv_facade
        yield Right(self.missing_in(venv_facade))

    @property
    def _str_extra(self) -> List[Any]:
        return List(self.plugins, self.venvs)

    @property
    def venv_dir(self) -> Either[str, Path]:
        return self.vim.vars.ppath('venv_dir').o(_default_venv_dir)

    @property
    def venv_facade(self) -> Either[str, VenvFacade]:
        return self.venv_dir / VenvFacade

    @property
    def want_init(self) -> Boolean:
        return self.vim.vars.pb('autostart') | true

    @property
    def inactive(self) -> List[Venv]:
        return self.installed.remove_all(self.active)

    def activate_venv(self, venv: Venv) -> 'Env':
        return self.append1.active(venv)

    def installed_by_name(self, names: List[str]) -> List[Venv]:
        return self.installed.filter(lambda v: v.name in names)

__all__ = ('Env',)



from chromatin.state import ChromatinTransitions, ChromatinComponent
from chromatin.plugins.core.messages import (AddPlugin, ShowPlugins, StageI, SetupPlugins, SetupVenvs, InstallMissing,
                                             AddVenv, IsInstalled, Activated, StageII, PostSetup, Installed,
                                             UpdatePlugins, Updated, Reboot, Activate)
from chromatin.venvs import VenvExistent
from chromatin.env import Env
from chromatin.venv import Venv
from chromatin.logging import Logging
from chromatin.host import PluginHost
from chromatin.util import resources

from amino.state import State, EitherState
from amino import __, do, _, Either, L, Just, Future, List, Right, Boolean
from amino.util.string import camelcaseify
from amino.boolean import true, false

from ribosome.machine import Message
from ribosome.machine.base import io, RunIOsParallel, SubProcessSync
from ribosome.machine.transition import Error
from ribosome.machine import trans
from ribosome.nvim import NvimFacade
from ribosome.machine.messages import Info
from ribosome.rpc import define_handler, RpcHandlerSpec


class PluginFunctions(Logging):

    @do
    def setup_venvs(self) -> EitherState[Env, List[Message]]:
        '''check whether a venv exists for each plugin in the self.env.
        for those that don't, create self.venvs in `g:chromatin_venv_dir`.
        '''
        venv_facade = yield EitherState.inspect_f(_.venv_facade)
        plugins = yield EitherState.inspect(_.plugins)
        jobs = plugins / venv_facade.check
        existent, absent = jobs.split_type(VenvExistent)
        ios = absent / _.plugin / venv_facade.bootstrap / __.map(AddVenv).map(_.pub)
        yield EitherState.pure(existent.map(_.venv).map(IsInstalled).cat(RunIOsParallel(ios)))

    @do
    def install_plugins(self, venvs: List[Venv], update: Boolean) -> EitherState[Env, List[Message]]:
        '''run subprocesses in sequence that install packages into their venvs using pip.
        cannot be run in parallel as they seem to deadlock.
        '''
        Done = update.cata(Updated, Installed)
        def trans_result(venv: Venv, result: Future) -> Message:
            return Just((Done(venv) if result.success else Error(result.err)).pub)
        venv_facade = yield EitherState.inspect_f(_.venv_facade)
        msgs = venvs / (lambda v: SubProcessSync(venv_facade.install(v), L(trans_result)(v, _)))
        yield EitherState.pure(msgs)

    @do
    def install_missing(self) -> EitherState[Env, List[Message]]:
        missing = yield EitherState.inspect_f(_.missing)
        yield self.install_plugins(missing, false)

    @do
    def update_plugins(self, plugins: List[str]) -> EitherState[Env, List[Message]]:
        getter = _.installed if plugins.empty else __.installed_by_name(plugins)
        venvs = yield EitherState.inspect(getter)
        yield (
            EitherState.pure(List(Error(resources.no_plugins_match_for_update(plugins))))
            if venvs.empty else
            self.install_plugins(venvs, true)
        )

    @do
    def activate_venv(self, vim: NvimFacade, venv: Venv) -> Either[str, Activated]:
        host = PluginHost(vim)
        yield host.start(venv)
        yield host.require(venv)
        yield Right(Activated(venv))

    @do
    def activate_multi(self, venvs: List[Venv]) -> EitherState[Env, List[Message]]:
        vim = yield EitherState.inspect(_.vim)
        yield EitherState.pure(venvs / L(self.activate_venv)(vim, _) / __.value_or(Error))

    @do
    def activate_by_names(self, plugins: List[str]) -> EitherState[Env, List[Message]]:
        getter = _.installed if plugins.empty else __.installed_by_name(plugins)
        venvs = yield EitherState.inspect(getter)
        yield (
            EitherState.pure(List(Error(resources.no_plugins_match_for_activation(plugins))))
            if venvs.empty else
            self.activate_multi(venvs)
        )

    def activate_all(self) -> EitherState[Env, List[Message]]:
        return self.activate_by_names(List())

    @do
    def activate_newly_installed(self) -> EitherState[Env, List[Message]]:
        new = yield EitherState.inspect(_.inactive)
        yield self.activate_multi(new)


class CoreTransitions(ChromatinTransitions):

    @property
    def funcs(self) -> PluginFunctions:
        return PluginFunctions()

    @trans.one(StageI)
    def stage_i(self) -> Message:
        return io(__.vars.set_p('started', True))

    @trans.multi(StageII, trans.st)
    @do
    def stage_ii(self) -> State[Env, List[Message]]:
        init = yield State.inspect(_.want_init)
        yield State.pure(List(SetupPlugins()) if init else List())

    @trans.unit(AddPlugin, trans.st)
    def add_plugin(self) -> State[Env, None]:
        spec = self.msg.spec
        name = self.msg.options.get('name') | spec
        return State.modify(__.add_plugin(name, spec))

    @trans.one(ShowPlugins, trans.st)
    @do
    def show_plugins(self) -> State[Env, Message]:
        venv_dir = yield State.inspect(_.venv_dir)
        venv_dir_msg = f'virtualenv dir: {venv_dir.value}'
        yield State.pure(Info(self.data.show_plugins.cons(venv_dir_msg).join_lines))

    @trans.multi(SetupPlugins)
    def setup_plugins(self) -> List[Message]:
        return List(SetupVenvs(), InstallMissing().at(.9).pub, PostSetup().at(.95).pub)

    @trans.multi(SetupVenvs, trans.est)
    def setup_venvs(self) -> EitherState[Env, List[Message]]:
        return self.funcs.setup_venvs()

    @trans.multi(InstallMissing, trans.est)
    def install_missing(self) -> EitherState[Env, List[Message]]:
        return self.funcs.install_missing()

    @trans.multi(Installed, trans.st)
    def installed(self) -> State[Env, List[Message]]:
        venv = self.msg.venv
        return State.pure(List(IsInstalled(venv), Info(resources.installed_plugin(venv.name))))

    @trans.multi(Updated, trans.st)
    def updated(self) -> State[Env, List[Message]]:
        venv = self.msg.venv
        return State.pure(List(Reboot(venv), Info(resources.updated_plugin(venv.name))))

    @trans.unit(IsInstalled, trans.st)
    def is_installed(self) -> State[Env, None]:
        return State.modify(__.add_installed(self.msg.venv))

    @trans.multi(PostSetup, trans.est)
    def post_setup(self) -> EitherState[Env, List[Message]]:
        return self.funcs.activate_newly_installed()

    @trans.unit(AddVenv, trans.st)
    def add_venv(self) -> State[Env, None]:
        return State.modify(__.add_venv(self.msg.venv))

    @trans.multi(Activate, trans.est)
    def activate(self) -> EitherState[Env, List[Message]]:
        return self.funcs.activate_all(self.msg.plugins)

    @trans.unit(Activated, trans.st)
    def activated(self) -> State[Env, None]:
        self.log.debug(f'activated {self.msg.venv}')
        venv = self.msg.venv
        cname = camelcaseify(venv.name)
        spec = RpcHandlerSpec.cmd(1, f'{cname}SetupRpc', dict())
        define_handler(self.vim, venv.name, spec, venv.plugin_path)
        self.vim.cmd_sync(f'{cname}SetupRpc', verbose=True)
        self.vim.runtime(f'chromatin/{venv.name}/*')
        return State.modify(__.activate_venv(venv))

    @trans.multi(UpdatePlugins, trans.est)
    def update_plugins(self) -> State[Env, List[Message]]:
        return self.funcs.update_plugins(self.msg.plugins)


class Plugin(ChromatinComponent):
    Transitions = CoreTransitions

__all__ = ('Plugin',)
